import numpy as np
from collections import OrderedDict

from ...Operation import Operation
from ... import optools
from . import saxstools

class SpectrumProfiler(Operation):
    """
    This operation profiles a SAXS spectrum (I(q) vs. q)
    to determine some characteristics of the sample.
    Based on some measures of the overall shape of the spectrum,
    the spectrum is tested for scattering from precursors 
    (approximated as small dilute monodisperse spheres),
    scattering from dilute nanoparticles of various form factors 
    (currently only spheres),
    diffraction peaks
    (Voigt-like profiles due to crystalline arrangements),
    or some combination of the three.
    TODO: document algorithm here.

    Output a return code
    and a dictionary of the results.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q', 'I', 'dI']
        output_names = ['return_code','features']
        super(SpectrumProfiler, self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['dI'] = 'optional- 1d array of intensity uncertainty values I(q)'
        self.output_doc['return_code'] = str('integer indicating '
        + 'whether or not the profiling finished. '
        + 'Possible values: 0=success, 1=error (raised an Exception), '
        + '2=warning (input data has unexpected characteristics).')
        self.output_doc['features'] = str('dict profiling the input spectrum. '
        + 'Dict keys and descriptions: \n'
        + 'WARNING_MESSAGE: warning message, only added if return_code==2. \n'
        + 'bad_data_flag: Boolean indicating that the spectrum is unfamiliar or mostly made of noise. \n'
        + 'precursor_flag: Boolean indicating presence of precursor terms. \n'
        + 'form_flag: Boolean indicating presence of form factor terms. \n'
        + 'structure_flag: Boolean indicating presence of structure factor terms. \n'
        + 'structure_id: Structure identity, e.g. fcc, hcp, bcc, etc. \n'
        + 'form_id: Form factor identity, e.g. sphere, cube, rod, etc. \n'
        + 'high_freq_ratio: Ratio of the upper half to the lower half '
        + 'of the power spectrum of the discrete fourier transform of the intensity. \n'
        + 'fluctuations_over_mean: Integrated fluctuation of intensity '
        + '(sum of absolute value of differences between adjacent points) '
        + 'divided by the mean intensity. \n'
        + 'bin_strengths: Integrated intensity in 10 evenly spaced q-bins. \n'
        + 'Imax_over_Imean: maximum intensity divided by mean intensity. \n'
        + 'Imax_over_Ilowq: maximum intensity divided by average intensity over lowest 10% of q domain. \n'
        + 'q_Imax: q value of the maximum intensity. \n'
        + 'q_corr: Pearson correlation between I(q) and q. \n'
        + 'qsquared_corr: Pearson correlation between I(q) and q^2. \n'
        + 'cos2q_corr: Pearson correlation between I(q) and (cos(q))^2. \n'
        + 'cosq_corr: Pearson correlation between I(q) and cos(q). \n'
        + 'invq_corr: Pearson correlation between I(q) and 1/q. \n'
        + 'q_logcorr: Same as q_corr, but with log(I(q)). \n'
        + 'qsquared_logcorr: Same as qsquared_corr, but with log(I(q)). \n'
        + 'cos2q_logcorr: Same as cos2q_corr, but with log(I(q)). \n'
        + 'cosq_logcorr: Same as cosq_corr, but with log(I(q)). \n'
        + 'invq_logcorr: Same as invq_corr, but with log(I(q)).')
        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['dI'] = optools.no_input
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type

    def run(self):
        q, I = self.inputs['q'], self.inputs['I']
        dI = self.inputs['dI']
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0
        self.outputs['return_code'] = 1

        ### amplitude analysis:
        idxmax = np.argmax(I)
        Imax = I[idxmax] 
        q_Imax = q[idxmax]
        Imean = np.mean(I)

        ### fourier analysis
        n_q = len(q)
        fftI = np.fft.fft(I)
        ampI = np.abs(fftI)
        powI = np.abs(fftI)**2
        high_freq_ratio = np.sum(powI[n_q/4:n_q/2])/np.sum(powI[1:n_q/4])
        #high_freq_ratio = np.sum(ampI[n_q/4:n_q/2])/np.sum(ampI[1:n_q/4])

        ### fluctuation analysis
        fluc = np.sum(np.abs(I[1:]-I[:-1]))
        fluctuations_over_mean = fluc/Imean 

	    # correlations on intensity
        q_corr = saxstools.compute_pearson(I,np.linspace(0,1,n_q))
        qsquared_corr = saxstools.compute_pearson(I,np.linspace(0,1,n_q)**2)
        cosq_corr = saxstools.compute_pearson(I,np.cos(q*np.pi/(2*q[-1])))
        cos2q_corr = saxstools.compute_pearson(I,np.cos(q*np.pi/(2*q[-1]))**2)
        invq_corr = saxstools.compute_pearson(I,q**-1)
        #invq4_corr = saxstools.compute_pearson(I,q**-4)

        # correlations on log intensity
        idx_nz = np.invert(I <= 0)
        q_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , np.linspace(0,1,n_q)[idx_nz] )
        qsquared_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , (np.linspace(0,1,n_q)[idx_nz])**2 )
        cosq_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , np.cos(q*np.pi/(2*q[-1]))[idx_nz] )
        cos2q_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , (np.cos(q*np.pi/(2*q[-1]))[idx_nz])**2 )
        invq_logcorr = saxstools.compute_pearson( np.log(I[idx_nz]) , q[idx_nz]**-1 )

        ### bin-integrated intensity analysis
        # TODO: standardize these bins in q space
        qmin, qmax = q[0], q[-1]
        qrange = qmax-qmin
        bin_strengths = np.zeros(10)
        for i in range(10):
            qmini, qmaxi = qmin+i*qrange/10,qmin+(i+1)*qrange/10
            idxi = list((q>=qmini) & (q<qmaxi))
            idxi_shift = [False]+idxi[:-1]
            idxi = np.array(idxi,dtype=bool)
            idxi_shift = np.array(idxi_shift,dtype=bool)
            dqi = q[ idxi_shift ] - q[ idxi ]
            Ii = I[ idxi ]
            bin_strengths[i] = np.sum(Ii * dqi) / (qmaxi-qmini)
        low_q_dominance = np.sum(bin_strengths[:5])/np.sum(bin_strengths[5:10]) 

        # Flagging bad data: 
        # Bad data have high noise
        # (high fluctuations relative to the mean)
        # or may be increasing-ish in q
        bad_data_flag = (fluctuations_over_mean > 30 
                        or q_corr > 0.2
                        or low_q_dominance < 2)
        Imax_over_Ilowq = float(Imax)/bin_strengths[0]
        Imax_over_Imean = float(Imax)/float(Imean)
        form_id = None 
        structure_id = None 
        if bad_data_flag:
            form_flag = False
            precursor_flag = False
            structure_flag = False
        else:
            # Flagging form factor: These tend to (log-)correlate with 1/q.
            # Spectra with relatively small form factor contribution 
            # tend to linear-correlate strongly with 1/q
            form_flag = ((invq_logcorr > 0.7 or invq_corr > 0.95) 
                        and low_q_dominance > 5 
                        and bin_strengths[0] / bin_strengths[-1] > 100)
            #and not all([cosq_corr > 0.7,cosq_logcorr > 0.7, cos2q_logcorr > 0.7, cos2q_corr>0.7]))
            if form_flag:
                # TODO: determine form factor here
                form_id = 'sphere'
 
            # Flagging precursors: 
            # Bin strengths should be decreasing, but not wildly  
            precursor_flag = low_q_dominance > 2 and low_q_dominance < 1000 

            # Flagging structure:
            # Largest bin is likely not lowest-q bin 
            structure_flag = Imax_over_Ilowq > 1.2 and q_Imax > 0.05 #and Imax_over_Imean > 4 
            # or (not np.argmax(bin_strengths) == 0))
            if structure_flag:
                # TODO: determine structure factor here
                structure_id = 'fcc'
        #######
        # More good/bad filters could be added here.
        #flag_msg = ''
        #if q_corr > 0.2:
        #    ok_flag = False
        #flag_msg = str('This algorithm expects spectra that tend to decrease in q. '
        #+ 'The input spectrum seems to be overall increasing in q.')
        #if (...) 
        #    ok_flag = False
        #    flag_msg = 'This spectrum seems mostly flat.'
        #elif (...):
        #    ok_flag = False
        #    flag_msg = ''
        #######
        ok_flag = True
        d_r = OrderedDict() 
        if not ok_flag:
            d_r['WARNING_MESSAGE'] = flag_msg
            self.outputs['return_code'] = 1
        else:
            d_r['bin_strengths'] = bin_strengths
            d_r['Imax_over_Imean'] = Imax_over_Imean
            d_r['Imax_over_Ilowq'] = Imax_over_Ilowq
            d_r['low_q_dominance'] = low_q_dominance 
            d_r['high_freq_ratio'] = high_freq_ratio 
            d_r['fluctuations_over_mean'] = fluctuations_over_mean 
            d_r['q_Imax'] = q_Imax
            d_r['q_logcorr'] = q_logcorr
            d_r['qsquared_logcorr'] = qsquared_logcorr
            d_r['cos2q_logcorr'] = cos2q_logcorr
            d_r['cosq_logcorr'] = cosq_logcorr
            d_r['invq_logcorr'] = invq_logcorr
            d_r['q_corr'] = q_corr
            d_r['qsquared_corr'] = qsquared_corr
            d_r['cos2q_corr'] = cos2q_corr
            d_r['cosq_corr'] = cosq_corr
            d_r['invq_corr'] = invq_corr
            d_r['bad_data_flag'] = bad_data_flag
            d_r['precursor_flag'] = precursor_flag
            d_r['form_flag'] = form_flag
            d_r['structure_flag'] = structure_flag
            d_r['structure_id'] = structure_id 
            d_r['form_id'] = form_id 
        self.outputs['return_code'] = 0
        self.outputs['features'] = d_r 

        #low_q_idxs = (q < q[0]+0.1*(q[-1]-q[0]))
        #high_q_idxs = (q > q[0]+0.9*(q[-1]-q[0]))
        #n_low_q = np.sum(np.array(low_q_idxs))
        #n_high_q = np.sum(np.array(high_q_idxs))
        # If the max intensity is outside the low-q region, throw flag
        #if not np.argmax(I) in range(2*n_low_q):
        #    ok_flag = False
        #    flag_msg = str('The maximum intensity '
        #    + '({} at q={}) '.format(np.max(I),q[np.argmax(I)])
        #    + 'is outside the lower 20% of the q-range '
        #    + '({} to {}).'.format(q[0],q[2*n_low_q]))
        # If there is a huge maximum somewhere, throw flag
        #elif np.max(I) > 10*np.mean(I[:2*n_low_q]):
        #    ok_flag = False
        #    flag_msg = str('The maximum intensity '
        #    + '({} at q={}) '.format(np.max(I),q[np.argmax(I)])
        #    + 'is greater than 10x the mean intensity '
        #    + 'of the lower 20% of the q-range '
        #    + '({} from {} to {}).'
        #    .format(np.mean(I[low_q_idxs]),q[0],q[n_low_q]))
        # If high-q intensity not significantly smaller than low-q, throw flag
        #elif not np.sum(I[:n_low_q]) > 10*np.sum(I[-1*n_high_q:]):
        #    ok_flag = False
        #    flag_msg = str('The lower 10% of the q-range '
        #    + '({} to {}) '.format(q[0],q[n_low_q])
        #    + 'has less than 10 times the intensity '
        #    + 'of the upper 10% of the q-range '
        #    + '({} to {}).'.format(q[-1*n_high_q],q[-1]))





"""
Created on Mon Jun 06 18:02:32 2016

author(s): Fang Ren, Apurva Mehta
Module originally contributed by Fang Ren.
For details, refer to the recent paper submitted to ACS Combinatorial Science.
TODO: Get this citation
"""

import numpy as np

from ...Operation import Operation
from ... import optools

class IntensityFeatures(Operation):
    """
    Extract the maximum intensity, average intensity, and a ratio of the two from data
    """

    def __init__(self):
        input_names = ['I']
        output_names = ['Imax', 'Iave', 'Imax_Iave_ratio']
        super(IntensityFeatures, self).__init__(input_names, output_names)
        self.input_doc['I'] = 'A 1d vector representing the intensity spectrum'
        self.input_src['I'] = optools.wf_input
        self.output_doc['Imax'] = 'The maximum intensity '
        self.output_doc['Iave'] = 'The average intensity'
        self.output_doc['Imax_Iave_ratio'] = 'The ratio of maximum to average intensity'

    def run(self):
        Imax = np.max(self.inputs['I'])
        Iave = np.mean(self.inputs['I'])
        ratio = Imax/Iave
        self.outputs['Imax'] = Imax
        self.outputs['Iave'] = Iave
        self.outputs['Imax_Iave_ratio'] = ratio




import numpy as np
import pyFAI

from ...Operation import Operation
from ... import optools

class ReduceByWXDDict_mask_error(Operation):
    """
    Input image data (ndarray) and a dict of calibration parameters 
    Return q, I(q)
    """
    def __init__(self):
        input_names = ['image_data','cal_params','pixel_size','fpolz','calculate_noise','inverse_gain','readnoise','mask']
        output_names = ['q','I_of_q','dI_of_q']
        super(ReduceByWXDDict_mask_error,self).__init__(input_names,output_names)
        # docs
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['cal_params'] = str( 'dict of calibration parameters read in from a .calib file,'
        + ' with keys d_pixel, rotation_rad, tilt_rad, lambda, x0_pixel, y0_pixel,'
        + ' PP (polarization factor), pixel_size, and d_pixel')
        self.input_doc['pixel_size'] = 'pixel size in microns'
        self.input_doc['fpolz'] = 'polarization factor'
        self.input_doc['calculate_noise'] = 'Calculate error estimate?'
        self.input_doc['inverse_gain'] = str('inverse gain, i.e. electrons on chip per digital count;'
        + ' applies primarily to CCDs')
        self.input_doc['readnoise'] = 'detector readnoise in electrons; applies primarily to CCDs'
        self.input_doc['mask'] = str('Bad pixel locations, if known (e.g., inactive detector regions, dead pixels,'
        + ' zingers/cosmic rays).  Boolean array of same shape as image, True indicating'
        + ' bad data and False indicating good, or None for default behavior.'
        + ' Default masking behavior is to mask all non-positive pixels.')
        self.output_doc['q'] = 'Scattering vector magnitude q'
        self.output_doc['I_of_q'] = 'Integrated intensity at q'
        self.output_doc['dI_of_q'] = 'Error estimate of integrated intensity at q'
        # src & type
        self.input_src['image_data'] = optools.wf_input
        self.input_src['cal_params'] = optools.wf_input
        self.input_src['pixel_size'] = optools.text_input
        self.input_src['fpolz'] = optools.text_input
        self.input_src['calculate_noise'] = optools.text_input
        self.input_src['inverse_gain'] = optools.text_input
        self.input_src['readnoise'] = optools.text_input
        self.input_src['mask'] = optools.text_input
        self.input_type['image_data'] = optools.ref_type
        self.input_type['cal_params'] = optools.ref_type
        self.input_type['pixel_size'] = optools.float_type
        self.input_type['fpolz'] = optools.float_type
        self.input_type['calculate_noise'] = optools.bool_type
        self.input_type['inverse_gain'] = optools.float_type
        self.input_type['readnoise'] = optools.float_type
        # defaults
        self.inputs['pixel_size'] = 79.
        self.inputs['fpolz'] = 0.95
        self.inputs['calculate_noise'] = False
        self.inputs['inverse_gain'] = 5.5 # 1 for no gain, sort of
        self.inputs['readnoise'] = 9.0 # 0 for no readnoise
        self.inputs['mask'] = None

    def run(self):
        img = self.inputs['image_data']
        # initialization parameters, change into Fit2D format
        pxsz = self.inputs['pixel_size'] # in microns
        fpolz = self.inputs['fpolz']
        l = self.inputs['cal_params']['lambda'] # wavelength (in Angstrom)
        d = self.inputs['cal_params']['d_pixel']*pxsz*0.001 # converting distance from pixel-widths to millimeters
        rot = (2*np.pi-self.inputs['cal_params']['rotation_rad'])/(2*np.pi)*360
        tilt = self.inputs['cal_params']['tilt_rad']/(2*np.pi)*360
        x0 = self.inputs['cal_params']['x0_pixel']
        y0 = self.inputs['cal_params']['y0_pixel']
        # PyFAI magic go!
        p = pyFAI.AzimuthalIntegrator(wavelength=l)
        p.setFit2D(d,x0,y0,tilt,rot,pxsz,pxsz)
        # define detector mask, to screen bad pixels
        # 1 for masked pixels, and 0 for valid pixels
        if self.inputs['mask'] is None:
            detector_mask = np.ones(img.shape)*(img <= 0)
        else:
            detector_mask = self.inputs['mask']
        q, I_of_q = p.integrate1d(img, 1000, mask=detector_mask, polarization_factor=fpolz) # 2nd arg number of bins
        q = q * 1E9 
        # noise model
        if bool(self.inputs['calculate_noise']):
            inverse_gain = self.inputs['inverse_gain']
            readnoise = self.inputs['readnoise']
            noise_model_squared = inverse_gain**-2 * (inverse_gain * img + readnoise**2)
            _, dI_squared = p.integrate1d(noise_model_squared, 1000, mask=detector_mask, polarization_factor=fpolz)
            dI_of_q = dI_squared**0.5
        else:
            dI_of_q = None
        # output
        self.outputs['q'] = q
        self.outputs['I_of_q'] = I_of_q
        self.outputs['dI_of_q'] = dI_of_q






from __future__ import print_function
from collections import OrderedDict
import copy
import traceback

#from ..plugins.WorkflowPlugin import WorkflowPlugin
from .Workflow import Workflow
from ..operations.Operation import Operation, Batch, Realtime        
from ..operations import optools        
from .. import pawstools

class WfManager(object):
    """
    Manager for paws Workflows. 
    Stores a list of Workflow objects, 
    performs operations on them.
    Keeps a reference to a PluginManager
    for access to PawsPlugins.
    """

    def __init__(self,plugin_manager):
        #super(WfManager,self).__init__()
        self.workflows = OrderedDict() 
        self.plugman = plugin_manager
        self.logmethod = print 

    #def __getitem__(self,key):
    #    if key in self.workflows.keys():
    #        return self.workflows[key]
    #    else:
    #        raise KeyError('[{}] WfManager does not recognize workflow name {}.'
    #        .format(__name__,key))

    def n_wf(self):
        return len(self.workflows)

    def write_log(self,msg):
        self.logmethod(msg)
        #if self.logmethod:
        #else:
        #    print('- '+pawstools.timestr()+': '+msg)

    def add_wf(self,wfname):
        """
        Add a workflow to self.workflows, with key specified by wfname.
        If wfname is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.
        """
        wf = Workflow()
        self.workflows[wfname] = wf

    def run_wf(self,wfname):
        """
        Serially execute the operations of WfManager.workflows[wfname].
        Uses optools.execution_stack() to determine execution order.
        """
        stk,diag = optools.execution_stack(self.workflows[wfname])
        for lst in stk:
            batch_flag = isinstance(self.workflows[wfname].get_data_from_uri(lst[0]),Batch)
            rt_flag = isinstance(self.workflows[wfname].get_data_from_uri(lst[0]),Realtime)
            if not any([batch_flag,rt_flag]):
                self.execute_serial(wfname,lst)
            elif batch_flag:
                self.execute_batch(wfname,lst[0],lst[1])
            elif rt_flag:
                self.execute_realtime(wfname,lst[0],lst[1])

    def execute_batch(self,wfname,batch_op_tag,batch_stk):
        batch_op = self.workflows[wfname].get_data_from_uri(batch_op_tag) 
        optools.load_inputs(batch_op,self.workflows[wfname],self.plugman)
        batch_op.run()
        self.workflows[wfname].set_item(batch_op_tag,batch_op)
        n_batch = len(batch_op.input_list())
        for i in range(n_batch):
            input_dict = batch_op.input_list()[i]
            for uri,val in input_dict.items():
                self.workflows[wfname].set_op_input_at_uri(uri,val)
            self.write_log( 'BATCH EXECUTION {} / {}'.format(i+1,n_batch) )
            for batch_lst in batch_stk:
                self.execute_serial(wfname,batch_lst)
            saved_items_dict = OrderedDict()
            for uri in batch_op.saved_items():
                # TODO # BUG: there is the chance for infinite recursion here
                # if the batch is asked to save an upstream item?
                save_data = copy.deepcopy(self.workflows[wfname].get_data_from_uri(uri))
                save_dict = self.uri_to_embedded_dict(uri,save_data) 
                saved_items_dict = self.update_embedded_dict(saved_items_dict,save_dict)
            #batch_op.output_list()[i] = copy.deepcopy(saved_items_dict)
            batch_op.output_list()[i] = saved_items_dict
            # TODO: set a more specific item here to save some tree update time?
            #self.workflows[wfname].set_item(batch_op_tag,batch_op)
            outputs_uri = batch_op_tag+'.'+optools.outputs_tag+'.'+batch_op.batch_outputs_tag()+'.'+str(i)
            self.workflows[wfname].set_item(outputs_uri,saved_items_dict)

    def execute_realtime(self,wfname,rt_op_tag,rt_stk):
        print('[{}] realtime execution needs review'.format(__name__))
        #rt_op = self.workflows[wfname].get_data_from_uri(rt_op_tag) 
        #optools.load_inputs(rt_op,self.workflows[wfname],self.plugman)
        #rt_op.run()
        #self.workflows[wfname].set_item(rt_op_tag,rt_op)
        #keep_running = True
        #n_exec = 0
        #while keep_running:
        ##TODO: ways to control the loop exit condition
        #    vals = rt_op.input_iter().next()
        #    if not None in vals:
        #        n_exec += 1
        #        wait_iter = 0
        #        inp_dict = OrderedDict( zip(rt_op.input_routes(), vals) )
        #        self.write_log( 'REALTIME EXECUTION {}'.format(n_exec))
        #        for uri,val in inp_dict.items():
        #            self.workflows[wfname].set_item(uri,val) 
        #        for rt_lst in rt_stk:
        #            self.execute_serial(wfname,rt_lst)
        #        saved_items_dict = OrderedDict()
        #        for uri in rt_op.saved_items():
        #            save_data = self.workflows[wfname].get_data_from_uri(uri)
        #            save_dict = self.uri_to_embedded_dict(uri,save_data) 
        #            saved_items_dict = self.update_embedded_dict(saved_items_dict,save_dict)
        #        rt_op.output_list().append(saved_items_dict)
        #        self.workflows[wfname].set_item(rt_op_tag,rt_op)
        #    else:
        #        if wait_iter == 0:
        #            self.write_log( 'Waiting for new inputs...' )
        #        wait_iter += 1 
        #        time.sleep(float(rt_op.delay())/1000.0)
        #    if wait_iter > 1000:
        #        self.write_log('Waited too long. Exiting...')
        #        keep_running = False

    def execute_serial(self,wfname,op_list):
        self.write_log('workflow {} running {}'.format(wfname,op_list))
        for op_tag in op_list: 
            op = self.workflows[wfname].get_data_from_uri(op_tag) 
            optools.load_inputs(op,self.workflows[wfname],self.plugman)
            try:
                op.run() 
            except Exception as ex:
                tb = traceback.format_exc()
                self.write_log(str('Operation {} of workflow {} threw an error. '
                + '\nMessage: {} \nTrace: {}').format(op_tag,wfname,ex.message,tb)) 
            self.workflows[wfname].set_item(op_tag,op)

    def uri_to_embedded_dict(self,uri,data=None):
        path = uri.split('.')
        endtag = path[-1]
        d = OrderedDict()
        d[endtag] = data
        for tag in path[:-1][::-1]:
            parent_d = OrderedDict()
            parent_d[tag] = d
            d = parent_d
        return d

    def update_embedded_dict(self,d,d_new):
        for k,v in d_new.items():
            if k in d.keys():
                if isinstance(d[k],dict) and isinstance(d_new[k],dict):
                    # embedded dicts: recurse
                    d[k] = self.update_embedded_dict(d[k],d_new[k])
                else:
                    # existing d[k] is not dict, and d_new[k] is dict: replace  
                    d[k] = d_new[k]
            else:
                # d[k] does not exist: insert
                d[k] = v
        return d

    def load_from_dict(self,wfname,opman,opdict):
        """
        Create a workflow with name wfname.
        If wfname is not unique, self.workflows[wfname] is overwritten.
        Input opdict specifies operation setup,
        where each item in opdict provides enough information
        to get an Operation from OpManager opman
        and set up its Operation.input_locators.
        """
        self.add_wf(wfname)
        for opname, op_setup in opdict.items():
            op = self.build_op_from_dict(op_setup,opman)
            if isinstance(op,Operation):
                self.workflows[wfname].set_item(opname,op)
            else:
                self.write_log('[{}] Failed to load {}.'.format(uri))

    def op_setup_dict(self,op):
        op_mod = op.__module__[op.__module__.find('operations'):]
        op_mod = op_mod[op_mod.find('.')+1:]
        dct = OrderedDict() 
        dct['op_module'] = op_mod
        inp_dct = OrderedDict() 
        for name in op.inputs.keys():
            il = op.input_locator[name]
            inp_dct[name] = {'src':copy.copy(il.src),'tp':copy.copy(il.tp),'val':copy.copy(il.val)}
        dct[optools.inputs_tag] = inp_dct 
        return dct

    def build_op_from_dict(self,op_setup,opman):
        op_uri = op_setup['op_module']
        op = opman.get_data_from_uri(op_uri)
        if issubclass(op,Operation):
            op = op()
            op.load_defaults()
            il_setup_dict = op_setup[optools.inputs_tag]
            for name in op.inputs.keys():
                if name in il_setup_dict.keys():
                    src = il_setup_dict[name]['src']
                    #if 'tp' in il_setup_dict[name].keys():
                    tp = il_setup_dict[name]['tp']
                    val = il_setup_dict[name]['val']
                    if tp in optools.invalid_types[src]:
                        il = optools.InputLocator(src,optools.none_type,None)
                    else:
                        il = optools.InputLocator(src,tp,val)
                    op.input_locator[name] = il
                    # dereference any existing inputs
                    # LAP: commented out bc it should be handled in op.load_defaults()
                    #op.inputs[name] = None
            return op
        else:
            return None

    #def auto_name(self,wfname):
    #    """
    #    Generate the next unique workflow name by appending '_x',
    #    where x is a minimal nonnegative integer.
    #    """
    #    goodname = False
    #    prefix = wfname
    #    idx = 1
    #    while not goodname:
    #        if not wfname in self.workflows.keys():
    #            goodname = True
    #        else:
    #            wfname = prefix+'_{}'.format(idx)
    #            idx += 1
    #    return wfname 





from collections import OrderedDict

from PySide import QtCore

from .QTreeSelectionModel import QTreeSelectionModel
from ..core.plugins.PawsPlugin import PawsPlugin
from ..core.operations import optools

class QPluginManager(QTreeSelectionModel):
    """
    A Qt Signal-slot manager for a TreeModel PluginManager.
    Takes a reference to a PluginManager in the constructor.
    The QPluginManager works mostly by
    calling on the methods of the PluginManager.
    """
    # TODO: Clean up headerData

    def __init__(self,plugman):
        flag_dict = OrderedDict()
        flag_dict['select'] = False
        super(QPluginManager,self).__init__(flag_dict,plugman)
        # plugman is accessible by self._tree after QTreeModel.__init__()
        #self.plugman = plugman
        

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} plugin(s) loaded".format(self._tree._root_item.n_children())
        else:
            return super(QPluginManager,self).headerData(section,orientation,data_role)    

    def add_plugin(self,pgin_tag,pgin):
        """Add a Plugin to the tree at the top level."""
        self.set_item(pgin_tag,pgin,self.root_index())

    def remove_plugin(self,pgin_tag):
        self.remove_item(pgin_tag)

    @QtCore.Slot(str)
    def update_plugin(self,pgin_name):
        #import pdb; pdb.set_trace()
        self.set_item(pgin_name,self._tree.get_data_from_uri(pgin_name))

    def load_from_dict(self,plugin_dict):
        """
        Load plugins from a dict that specifies their setup parameters.
        """
        for tag, pgin_spec in plugin_dict.items():
            pgin_uri = pgin_spec['plugin_module']
            pgin = self._tree.get_plugin(pgin_uri)
            if pgin is not None:
                if not issubclass(pgin,PawsPlugin):
                    self._tree.write_log('Did not find Plugin {} - skipping.'.format(pgin_uri))
                    return 
            pgin = pgin()
            for name in pgin.inputs.keys():
                if name in pgin_spec[optools.inputs_tag]:
                    pgin.inputs[name] = pgin_spec[optools.inputs_tag][name]
            pgin.start()
            self.add_plugin(tag,pgin)

    #        self.plugman.contains_uri(pgin_name):
    #        #pgin = self.get_data_from_uri(pgin_name)
    #        #self.tree_update(self.root_index(),pgin_name,self.build_tree(pgin))

    #def remove_plugin(self,rm_idx):
    #    """Remove a Plugin from the tree"""
    #    p_idx = self.parent(rm_idx)
    #    if not p_idx == self.root_index():
    #        msg = '[{}] Called remove_plugin on non-Plugin at QModelIndex {}. \n'.format(__name__,rm_idx)
    #        raise ValueError(msg)
    #    self.remove_item(rm_itm.tag,p_idx)

    # Overloaded headerData() for PluginManager 
    #def headerData(self,section,orientation,data_role):
    #    if (data_role == QtCore.Qt.DisplayRole and section == 0):
    #        return "Plugins: {} active".format(self.item_count())
    #    elif (data_role == QtCore.Qt.DisplayRole and section == 1):
    #        return super(PluginManager,self).headerData(section,orientation,data_role)    
    #    else:
    #        return None






from matplotlib import pyplot as plt

from ..Operation import Operation
from .. import optools

class MPLFig_2Curves(Operation):

    def __init__(self):
        input_names = ['x_y_1', 'x_y_2']
        output_names = ['figure']
        super(MPLFig_2Curves, self).__init__(input_names, output_names)
        self.input_doc['x_y_1'] = 'n-by-2 array of y versus x for curve 1'
        self.input_doc['x_y_2'] = 'n-by-2 array of y versus x for curve 2'
        self.output_doc['figure'] = 'matplotlib.pyplot.Figure with the two curves plotted together'
        self.input_src['x_y_1'] = optools.wf_input
        self.input_src['x_y_2'] = optools.wf_input
        self.input_type['x_y_1'] = optools.ref_type
        self.input_type['x_y_2'] = optools.ref_type

    def run(self):
        fig = plt.figure()
        x_y_1 = self.inputs['x_y_1']
        x_y_2 = self.inputs['x_y_2']
        plt.plot(x_y_1[:,0],x_y_1[:,1],'g')
        plt.plot(x_y_2[:,0],x_y_2[:,1],'r')
        self.outputs['figure'] = fig






from collections import OrderedDict
import glob

from ...Operation import Batch
from ... import optools

class BatchFromFiles(Batch):
    """
    Read a directory and filter its contents with a regular expression
    to form  a list of file paths to be used as inputs
    for the repeated execution of a workflow.
    Specify, by workflow uri, where this file path will be fed to the workflow.
    Collect specified outputs from the workflow for each of the inputs.
    """

    def __init__(self):
        input_names = ['dir_path','regex','input_route','batch_ops','saved_items']
        output_names = ['batch_inputs','batch_outputs']
        super(BatchFromFiles,self).__init__(input_names,output_names)
        self.input_doc['dir_path'] = 'path to directory containing batch of files to be used as input'
        self.input_doc['regex'] = 'string with * wildcards that will be substituted to indicate input files'
        self.input_doc['input_route'] = 'inputs constructed by the batch executor are directed to this uri'
        self.input_doc['batch_ops'] = str('list of workflow uris pointing to Operations to be included in batch execution- '
        + 'the order of entries is unimportant, as the proper execution stack is resolved at runtime')
        self.input_doc['saved_items'] = 'list of uris to be saved in the batch_outputs'
        self.output_doc['batch_inputs'] = 'list of dicts of [input_route:input_value]'
        self.output_doc['batch_outputs'] = 'list of dicts of [output_route:output_value] for all saved_items '
        self.input_src['dir_path'] = optools.fs_input
        self.input_src['regex'] = optools.text_input 
        self.input_src['input_route'] = optools.wf_input 
        self.input_src['batch_ops'] = optools.wf_input 
        self.input_src['saved_items'] = optools.wf_input 
        self.input_type['dir_path'] = optools.path_type
        self.input_type['regex'] = optools.str_type
        self.input_type['input_route'] = optools.path_type
        self.input_type['batch_ops'] = optools.path_type 
        self.input_type['saved_items'] = optools.path_type 
        self.inputs['regex'] = '*.tif' 
        self.inputs['batch_ops'] = []
        self.inputs['saved_items'] = []
        
    def run(self):
        """
        Build a list of [uri:value] dicts to be used in the workflow.
        """
        dirpath = self.inputs['dir_path']
        rx = self.inputs['regex']
        inproute = self.inputs['input_route']
        #batch_list = [dirpath+'/'+rx.replace('*',sub) for sub in subs]
        batch_list = glob.glob(dirpath+'/'+rx)
        input_dict_list = []
        output_dict_list = []
        for filename in batch_list:
            inp_dict = OrderedDict() 
            inp_dict[inproute] = filename
            input_dict_list.append(inp_dict)
            output_dict_list.append(OrderedDict())
        self.outputs['batch_inputs'] = input_dict_list
        # Instantiate the batch_outputs list
        self.outputs['batch_outputs'] = output_dict_list 

    def input_list(self):
        return self.outputs['batch_inputs']

    def output_list(self):
        return self.outputs['batch_outputs']

    def input_routes(self):
        """Provide the input route- a list is expected"""
        #return [self.input_locator['input_route'].val]
        if isinstance(self.inputs['input_route'],list):
            return self.inputs['input_route']
        else:
            return [self.inputs['input_route']]

    def batch_ops(self):
        """Provide a list of uri's of ops to be included in batch execution"""
        if isinstance(self.inputs['batch_ops'],list):
            return self.inputs['batch_ops']
        else:
            return [self.inputs['batch_ops']]

    def saved_items(self):
        """List uris to be saved/stored after execution"""
        if isinstance(self.inputs['saved_items'],list):
            return self.inputs['saved_items']
        else:
            return [self.inputs['saved_items']]

    def batch_outputs_tag(self):
        return 'batch_outputs'

    def set_batch_ops(self,wf=None):
        self.inputs['batch_ops'] = optools.locate_input(self.input_locator['batch_ops'],wf)

    def set_input_routes(self,wf=None):
        self.inputs['input_route'] = optools.locate_input(self.input_locator['input_route'],wf)







from collections import OrderedDict

from PySide import QtCore

from ..qt.QTreeSelectionModel import QTreeSelectionModel

class QOpManager(QTreeSelectionModel):
    """
    A QTreeSelectionModel for interacting with TreeModel OpManager.
    """

    def __init__(self,opman):
        flag_dict = OrderedDict()
        flag_dict['enable'] = True
        super(QOpManager,self).__init__(flag_dict,opman)
        self.opman = opman

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} operations available".format(self.opman.n_ops())
        else:
            return super(QOpManager,self).headerData(section,orientation,data_role) 

    #def idx_of_cat(self,catname,p_idx):
    #    """
    #    If cat exists under p_idx, return its index, 
    #    else return an invalid QModelIndex.
    #    """
    #    ncats = self.item_count(p_idx)
    #    p_itm = self.get_from_idx(p_idx)
    #    for j in range(ncats):
    #        idx = self.index(j,0,p_idx)
    #        cat = self.get_from_idx(idx).tag
    #        if cat == catname:
    #            return idx
    #    return QtCore.QModelIndex() 

    #def remove_ops(self,rm_idx):
    #    """
    #    Remove an Operation or category from the tree.
    #    """
    #    if not rm_idx == self.root_index() and rm_idx.isValid():
    #        p_idx = self.parent(rm_idx)
    #        rm_itm = self.get_from_idx(rm_idx)
    #        self.delete_item(rm_itm.tag,p_idx)            

    # get an Operation by its cat.[subcat.(...).]opname uri
    #def get_op_from_uri(self,op_uri):
        #return self.get_data_from_uri(op_uri)
        #for op in self._op_list:
        #    if op.__name__ == op_name:
        #        return op
        #return None

    # get an Operation from the list by its TreeItem's QModelIndex
    #def get_op(self,indx):
        #treeitem = self.get_item(indx)
        #return treeitem.data
 





"""
Created on Mon Jun 06 18:02:32 2016

author(s): Fang Ren, Apurva Mehta
Module originally contributed by Fang Ren.
For details, refer to the recent paper submitted to ACS Combinatorial Science.
TODO: get this citation 
"""

import numpy as np

from ...Operation import Operation
from ... import optools

class FindLocalMaxima(Operation):
    """
    Extract the locations and intensities of local maxima in a 1D spectrum.
    TODO: Document the algorithm here.
    """

    def __init__(self):
        input_names = ['q', 'I', 'delta_I']
        output_names = ['q_pk', 'I_pk']
        super(PeakFeatures,self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array for x-axis of spectrum, named q for momentum transfer vector'
        self.input_doc['I'] = '1d array for spectral intensities at q values'
        self.input_doc['delta_I'] = str('Criterion for peak finding: point is a maximum '
            + 'if it is more than delta_I larger than the next point')
        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['delta_I'] = optools.text_input
        self.input_type['delta_I'] = optools.float_type
        self.inputs['delta_I'] = 0.0
        self.output_doc['q_pk'] = 'q values of found peaks'
        self.output_doc['I_pk'] = 'intensities of found peaks'

    def run(self):
        maxtab, mintab = self.get_extrema(self.inputs['q'], self.inputs['I'], self.inputs['delta_I'])
        q_pk = maxtab[:,0]
        I_pk = maxtab[:,1] 
        # save results to self.outputs
        self.outputs['q_pk'] = q_pk 
        self.outputs['I_pk'] = I_pk 

    def get_extrema(x, y, delta):
        """
        Given vectors x and y, return an n-by-2 array of x,y pairs 
        for the minima and maxima of y(x)
        """
        maxtab = []
        mintab = []
        x = np.array(x)
        y = np.array(y)
        y_min, y_max = np.Inf, -np.Inf
        x_min, x_max = np.NaN, np.NaN
        for i in arange(len(y)):
            if y[i] > y_max:
                x_max = x[i]
                y_max = y[i] 
            if y[i] < y_min:
                x_min = x[i]
                y_min = y[i] 
            lookformax = True
            if lookformax:
                if y[i] < y_max-delta:
                    maxtab.append((x_max, y_max))
                    y_min = y[i] 
                    x_min = x[i]
                    lookformax = False
            else:
                if y[i] > y_min+delta:
                    mintab.append((x_min, y_min))
                    y_max = y[i] 
                    x_max = x[i]
                    lookformax = True
        return np.array(maxtab), np.array(mintab)





import numpy as np

from ...Operation import Operation
from ...import optools

class CSVToArray(Operation):
    """
    Read a csv-formatted file into a numpy array.
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['array']
        super(CSVToArray, self).__init__(input_names, output_names)
        self.input_doc['path'] = "path to .csv file"
        self.output_doc['array'] = "numpy array built from csv file contents"
        # source & type
        self.input_src['path'] = optools.fs_input
        self.input_type['path'] = optools.path_type

    def run(self):
        path = self.inputs['path']
        self.outputs['array'] = np.loadtxt(path, dtype=float, delimiter=',')





from .PawsPlugin import PawsPlugin
from ..workflow.WfManager import WfManager 

class WfManagerPlugin(PawsPlugin):
    """
    This plugin exposes the content of the workflow manager.
    """

    def __init__(self):
        input_names = ['wf_manager']
        super(WfManagerPlugin,self).__init__(input_names)
        self.input_doc['wf_manager'] = 'A workflow manager (paws.core.workflow.WfManager.WfManager).'

    def start(self):
        self.wfman = self.inputs['wf_manager'] 
        pass 

    def stop(self):
        pass

    def content(self):
        if isinstance(self.wfman,WfManager):
            return self.wfman.workflows
        else:
            return {} 

    def description(self):
        return str('Workflow manager Plugin for paws: '
            + 'This wraps the workflow manager in a plugin, '
            + 'so that workflows can use the plugin '
            + 'to fetch data from other workflows.')





"""This module defines a class that presents an API for paws."""

import os
from functools import partial
from collections import OrderedDict

from ..core import pawstools
from ..core import operations as ops
from ..core.operations import optools
from ..core.operations.OpManager import OpManager 
from ..core.workflow.WfManager import WfManager 
from ..core.plugins.PluginManager import PluginManager 
from ..core.plugins.WfManagerPlugin import WfManagerPlugin

def start(app_args=[]):
    """
    Instantiate and return a PawsAPI object. 

    paws.api.start() calls the PawsAPI constructor.

    :param app_args: arguments to pass to the 
    QApplication constructor within the PawsAPI constructor
    :type app_args: sequence
    :returns: a PawsAPI object
    :return type: paws.api.PawsAPI 
    """
    return PawsAPI(app_args)

class PawsAPI(object):
    """
    A container to facilitate interaction with a set of paws objects:
    an Operations Manager, a Workflow Manager, and a Plugins Manager. 
    """

    def __init__(self,api_args):
        super(PawsAPI,self).__init__()
        self._op_manager = OpManager()
        self._plugin_manager = PluginManager()
        self._wf_manager = WfManager(self._plugin_manager)
        # TODO: load_cats and load_ops should happen outside the api.__init__
        # so that different api instances can have different operations loaded
        self._op_manager.load_cats(ops.cat_list) 
        self._op_manager.load_ops(ops.cat_op_list)
        self._current_wf_name = None 
        wfman_pgin = WfManagerPlugin()
        wfman_pgin.inputs['wf_manager'] = self._wf_manager 
        wfman_pgin.start()
        self._plugin_manager.set_item('wf_manager',wfman_pgin)
    
    def enable_ops(self,*args):
        # TODO: operation enable/disable functionality
        pass

    def add_wf(self,wfname):
        self._wf_manager.add_wf(wfname)
        if not self._current_wf_name:
            self.select_wf(wfname)

    def select_wf(self,wfname):
        if wfname in self._wf_manager.workflows.keys():
            self._current_wf_name = wfname
        else:
            msg = str('requested workflow {} not found in {}'
            .format(wfname,self._wf_manager.workflows.keys()))
            raise ValueError(msg)

    def current_wf(self):
        if self._current_wf_name:
            return self._wf_manager.workflows[self._current_wf_name]    
        else:
            return None

    def get_wf(self,wfname=None):
        if wfname is None:
            return self.current_wf()
        else:
            return self._wf_manager.workflows[wfname]

    def get_op(self,opname,wfname=None):
        return self.get_wf(wfname).get_data_from_uri(opname) 

    def add_op(self,op_tag,op_spec,wfname=None):
        wf = self.get_wf(wfname)
        # get the op referred to by op_spec
        op = self._op_manager.get_data_from_uri(op_spec)
        # instantiate with default inputs
        op = op()
        op.load_defaults()
        wf.set_item(op_tag,op)

    def add_plugin(self,pgin_tag,pgin_name):
        pgin = self._plugin_manager.get_plugin(pgin_name)
        pgin = pgin()
        self._plugin_manager.set_item(pgin_tag,pgin)

    def set_plugin_input(self,pgin_tag,input_name,**kwargs):
        src,tp,val = self._parse_src_tp_val(kwargs)
        if src is None:
            src = pgin.input_src[input_name]
        if tp is None:
            tp = pgin.input_type[input_name]
        if val is None:
            val = pgin.inputs[input_name]
        if src == optools.text_input:
            self.pgin.inputs[name] = optools.cast_type_val(tp,val)
        else:
            self.pgin.inputs[name] = val 

    def start_plugin(self,pgin_name):
        pgin = self.get_plugin(pgin_name)
        pgin.start()

    def get_plugin(self,pgin_name):
        return self._plugin_manager.get_data_from_uri(pgin_name)

    def remove_op(self,op_tag,wfname=None):
        wf = self.get_wf(wfname)
        wf.remove_item(op_tag)

    def set_input(self,opname,input_name,wfname=None,**kwargs):
        op = self.get_op(opname,wfname) 
        src,tp,val = self._parse_src_tp_val(kwargs)
        if src is None:
            src = op.input_locator[input_name].src
        if tp is None:
            tp = op.input_locator[input_name].tp
        if val is None:
            val = op.input_locator[input_name].val
        il = optools.InputLocator(src,tp,val)
        op.input_locator[input_name] = il

    @staticmethod
    def _parse_src_tp_val(kw_dict): 
        src = None
        tp = None
        val = None
        if 'src' in kw_dict:
            if kw_dict['src'] in optools.input_sources:
                src = optools.valid_sources[ optools.input_sources.index(kw_dict['src']) ]
                # Any default type setting can go here.
                if src == optools.batch_input:
                    tp = optools.auto_type
                if src == optools.fs_input:
                    tp = optools.path_type
                if src == optools.wf_input:
                    tp = optools.ref_type
            else:
                #TODO: error? warning?
                src = optools.no_input
        if 'tp' in kw_dict:
            if kw_dict['tp'] in optools.input_types:
                tp = optools.valid_types[ optools.input_types.index(kw_dict['tp']) ]
            else:
                #TODO: error? warning?
                tp = optools.none_type
        if 'val' in kw_dict:
            val = kw_dict['val']
        return src,tp,val

    def get_output(self,opname,output_name,wfname=None):
        op = self.get_op(opname,wfname)
        return op.outputs[output_name]

    def execute(self,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        self._wf_manager.run_wf(wfname)
        
    def save_config(self):
        ops.save_config()

    def save_workflow(self,wfl_filename):
        """
        Save the current workflow to a .wfl file,
        as specified by wfl_filename.
        If the given filename does not have the .wfl extension,
        it will be appended.
        """
        if not os.path.splitext(wfl_filename)[1] == '.wfl':
            wfl_filename = wfl_filename + '.wfl'
        self._wf_manager.logmethod( 'dumping workflow {} to {}'.format(self._current_wf_name,wfl_filename) )
        d = {} 
        wf_dict = OrderedDict() 
        for opname in self.current_wf().list_op_tags():
            op = self.current_wf().get_data_from_uri(opname)
            wf_dict[opname] = self._wf_manager.op_setup_dict(op)
        d['WORKFLOW'] = wf_dict
        pawstools.update_file(wfl_filename,d)

    def op_count(self,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        return self._wf_manager.workflows[wfname].n_children()

    def list_op_tags(self,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        return self._wf_manager.workflows[wfname].list_op_tags()

    #def inspect_objects(self):
    #    """
    #    Use QObject.findChildren() to count references to child objects
    #    of the top-level paws resource managers.
    #    Return a report of the result as a string.
    #    """
    #    opman_children = self._op_manager.findChildren(QtCore.QObject)
    #    wfman_children = self._wf_manager.findChildren(QtCore.QObject)
    #    plugman_children = self._plugin_manager.findChildren(QtCore.QObject)
    #    rpt = str('paws QObject count:\n'
    #        + 'operations manager: {}\n{}...\n'.format(len(opman_children),opman_children)
    #        + 'plugins manager: {}\n{}\n'.format(len(plugman_children),plugman_children)
    #        + 'workflow manager: {}\n{}\n'.format(len(wfman_children),wfman_children))
    #    for wfname,wf in self._wf_manager.workflows.items():
    #        wf_children = wf.findChildren(QtCore.QObject)
    #        rpt += '\tworkflow {}: {}\n\t{}\n'.format(wfname,len(wf_children),wf_children)
    #    return rpt
        
#def core_app(app_args=[]):
#    """
#    Return a reference to a new QCoreApplication or a currently running QApplication.
#    
#    Input arguments are passed to the QApplication constructor.
#    If a RuntimeError is thrown,
#    it is assumed that a QApplication is already running,
#    and an attempt is made to return a reference to that QApplication.
#    If that fails, this returns None.
#
#    :param app_args: arguments to pass to the QApplication constructor
#    :type args: sequence
#    :returns: reference to a new or existing QCoreApplication
#    :return type: PySide.QtCore.QCoreApplication or None
#    """
#    try:
#        app = QtCore.QCoreApplication(app_args)
#    except RuntimeError:
#        try:
#            app = QtCore.QCoreApplication.instance()
#        except:
#            app = None
#    return app




from functools import partial
from collections import OrderedDict

from PySide import QtGui, QtCore, QtUiTools

from .. import uitools
from .. import widgets
from ...core import pawstools
from ...core.operations import optools
from ...core.models.ListModel import ListModel

class PipelineWidget(QtGui.QWidget):
    """
    Stores a reference to a QMainWindow,
    performs operations on it
    """

    def __init__(self,paws_api):
        super(PipelineWidget,self).__init__()
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/pipeline_widget.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.paw = paws_api
        self._wfname = 'pipeline'
        self.paw.add_wf(self._wfname)
    
        # This is a dict to translate from on-screen operation names
        # to uris for locating the Operations in paws.
        self.ops = OrderedDict()
        self.ops['read_PONI'] = 'INPUT.CALIBRATION.ReadPONI'
        self.ops['read_background'] = 'INPUT.TIF.LoadTif'
        self.ops['read_image'] = 'INPUT.TIF.LoadTif'
        self.ops['calibrate_background'] = 'PROCESSING.CALIBRATION.Calibrate'
        self.ops['calibrate_image'] = 'PROCESSING.CALIBRATION.Calibrate'
        self.ops['subtract_background'] = 'PROCESSING.BACKGROUND.SubtractMaximumBackground'
        self.ops['saxs_profile'] = 'PROCESSING.SAXS.SAXSProfile'

        # This is a dict for declaring the dependencies of operations.
        # An operation should only be added to the pipeline
        # when all of its dependencies are already in there.
        self.deps = dict.fromkeys(self.ops.keys()) 
        self.deps['calibrate_background'] = ['read_background']
        self.deps['calibrate_image'] = ['read_image']
        self.deps['subtract_background'] = ['calibrate_background','calibrate_image']
        self.deps['saxs_profile'] = ['subtract_background']

        # This is a dict for specifying inputs that need to be set by the user
        # when an operation is loaded.
        self.required_inputs = dict.fromkeys(self.ops.keys())
        self.required_inputs['read_PONI'] = ['poni_file']
        self.required_inputs['read_background'] = ['path']
        self.required_inputs['read_image'] = ['path']

        # This is a dict for holding on to widgets,
        # so that they can be displayed in self.ui.output_tabs as needed.
        self.output_widgets = {} 

        self.build_ui()
        self.set_wf_widgets()

    def build_ui(self):
        self.ui.output_tabs.clear()
        self.ui.setWindowTitle('pipeline widget')
        self.ui.wf_groupBox.setTitle('workflow control')
        self.ui.output_groupBox.setTitle('output inspector')
        self.ui.run_wf_button.setText('&Run') 
        self.ui.run_wf_button.clicked.connect(self.run_wf) 
        self.ui.setStyleSheet( "QLineEdit { border: none }" + self.ui.styleSheet() )
        self.ui.splitter.setSizes([1,1])

    def set_wf_widgets(self):
        add_op_row = self.paw.op_count(self._wfname)+1
        self.ui.wf_layout.addWidget(self.add_op_button(),add_op_row,0,1,1)

    def add_op_button(self):
        btn = QtGui.QPushButton('&+...',self)
        btn.setMaximumWidth(60)
        # TODO: btn.setIcon(icon_image)
        btn.clicked.connect(self.add_op)
        return btn

    def add_op(self):
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/list_selector.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        add_op_ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        add_op_ui.selector_box.setTitle('select operation')
        add_op_ui.setWindowTitle('operation setup')
        op_list_model = ListModel(self.ops.keys())
        self.refresh_op_list(op_list_model)
        add_op_ui.op_list.setModel(op_list_model)
        add_op_ui.add_button.setText('&OK')
        add_op_ui.op_list.doubleClicked.connect( partial(self.finish_add_op,add_op_ui) )
        add_op_ui.add_button.clicked.connect( partial(self.finish_add_op,add_op_ui) )
        add_op_ui.done_button.setText('&Cancel')
        add_op_ui.done_button.clicked.connect(add_op_ui.close)
        add_op_ui.setParent(self.ui,QtCore.Qt.Window)
        add_op_ui.show()

    def finish_add_op(self,ui,op_idx=None):
        add_op_row = self.paw.op_count(self._wfname)+1
        if op_idx is None:
            # Get selection ui.op_list
            op_idx = ui.op_list.currentIndex()
        op_name = self.ops.keys()[op_idx.row()]
        op_tag = op_name
        if op_tag in self.paw.get_wf(self._wfname).list_op_tags():
            op_tag = self.paw.get_wf(self._wfname).make_unique_uri(op_name)
        op_uri = self.ops[op_name]
        # Add the op to the workflow
        # TODO: ensure op names are unique before wf loading.
        # TODO: then, make sure unique op names are still recognized upstream during automatic input loading.
        self.paw.add_op(op_tag,op_uri,self._wfname)
        # Set the default input structure
        self.default_op_setup(op_tag)
        # If op has inputs to set, whip out some widgets
        if self.required_inputs[op_name] is not None:
            for inpname in self.required_inputs[op_name]:
                print 'fetch {}.{}'.format(op_tag,inpname) 
                self.fetch_data(op_tag,inpname,ui)

        print 'proceed with op loading'
        # Add the op name in a text box
        op_tag_widget = QtGui.QLineEdit(op_tag)
        op_tag_widget.setReadOnly(True)
        self.ui.wf_layout.addWidget(op_tag_widget,add_op_row,0,1,1)
        
        # Add buttons to edit, remove, or toggle visualization
        op_edit_button = QtGui.QPushButton('edit')
        op_edit_button.clicked.connect( partial(self.edit_op,op_tag,op_name) )
        rm_op_button = QtGui.QPushButton('remove')
        rm_op_button.clicked.connect( partial(self.rm_op,op_tag) )
        vis_toggle = QtGui.QCheckBox('display')
        vis_toggle.stateChanged.connect( partial(self.set_visualizer,op_tag) )
        self.ui.wf_layout.addWidget(op_edit_button,add_op_row,1,1,1)
        self.ui.wf_layout.addWidget(rm_op_button,add_op_row,2,1,1)
        self.ui.wf_layout.addWidget(vis_toggle,add_op_row,3,1,1)

        # Refresh the ops list
        self.refresh_op_list(ui.op_list.model())
        ui.repaint()

        # Prepare workflow building widgets for whatever is next
        self.set_wf_widgets()
        print 'close add_op ui'
        ui.close()

    def edit_op(self,op_tag,op_name):
        for inpname in self.required_inputs[op_name]:
            self.fetch_data(op_tag,inpname,self.ui)
        #ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/op_quicksetup.ui")
        #ui_file.open(QtCore.QFile.ReadOnly)
        #op_setup_ui = QtUiTools.QUiLoader().load(ui_file)
        #ui_file.close()
        #op_setup_ui.inputs_box.setTitle('operation: {}'.format(op_tag))
        #op_setup_ui.setWindowTitle('operation setup')
        ## Get a reference to the Operation by name...
        ## TODO: this sort of op inspection should be rolled into the api
        #op = self.paw.get_op(op_tag)
        #val_widgets = {}
        #for inprow,inpnm in zip(range(len(op.inputs.keys())),op.inputs.keys()):
        #    inp_tag_widget = uitools.name_widget(inpnm)
        #    op_setup_ui.inputs_layout.addWidget(inp_tag_widget,inprow,0,1,1)
        #    eq_widget = uitools.smalltext_widget('=')
        #    op_setup_ui.inputs_layout.addWidget(eq_widget,inprow,1,1,1)
        #    val_widget = uitools.bigtext_widget(op.input_locator[inpnm].val) 
        #    val_widget.setReadOnly(False)
        #    op_setup_ui.inputs_layout.addWidget(val_widget,inprow,2,1,1)
        #    val_widgets[inpnm] = val_widget
        #    btn_widget = QtGui.QPushButton('browse...')
        #    btn_widget.setMaximumWidth(80)
        #    btn_widget.clicked.connect( partial(self.fetch_data,op_tag,inpnm,op_setup_ui) )
        #    inpsrc = optools.input_sources[op.input_locator[inpnm].src]
        #    if inpsrc == optools.text_input:
        #        btn_widget.setEnabled(False)
        #    op_setup_ui.inputs_layout.addWidget(btn_widget,inprow,3,1,1)
        #op_setup_ui.finish_button.setText('&OK')
        #op_setup_ui.finish_button.clicked.connect( partial(self.finish_edit_op,op_tag,val_widgets,op_setup_ui) )
        #op_setup_ui.setParent(self.ui,QtCore.Qt.Window)
        #op_setup_ui.show()

    def fetch_data(self,op_tag,input_name,parent_ui):
        print 'now fetching for {}.{}'.format(op_tag,input_name)
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/tree_selector.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        data_fetch_ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        data_fetch_ui.setParent(parent_ui,QtCore.Qt.Window)
        #data_fetch_ui.setWindowModality(QtCore.Qt.WindowModal)
        data_fetch_ui.setWindowModality(QtCore.Qt.ApplicationModal)
        data_fetch_ui.setWindowTitle('browse for {}.{}'.format(op_tag,input_name))
        op = self.paw.get_op(op_tag)
        src = op.input_locator[input_name].src 
        if src == optools.wf_input:
            # TODO: add a workflow selector?
            wf = self.paw.get_wf(self._wfname)
            data_fetch_ui.source_treeview.setModel(wf)
        elif src == optools.fs_input:
            fs_trmod = QtGui.QFileSystemModel()
            data_fetch_ui.source_treeview.setModel(fs_trmod)
            fs_trmod.setRootPath(QtCore.QDir.currentPath())
            idx = fs_trmod.index(QtCore.QDir.currentPath())
            while idx.isValid():
                data_fetch_ui.source_treeview.setExpanded(idx,True)
                idx = idx.parent()
            data_fetch_ui.source_treeview.setExpanded(fs_trmod.index(pawstools.sourcedir),True)
            data_fetch_ui.source_treeview.hideColumn(1)
            data_fetch_ui.source_treeview.hideColumn(3)
            data_fetch_ui.source_treeview.setColumnWidth(0,250)
        data_fetch_ui.finish_button.setText('&OK')
        data_fetch_ui.source_treeview.doubleClicked.connect( 
        partial(self.finish_fetch_data,op_tag,input_name,data_fetch_ui) )        
        data_fetch_ui.finish_button.clicked.connect( 
        partial(self.finish_fetch_data,op_tag,input_name,data_fetch_ui) )        
        print 'bringing up data_fetch ui'
        data_fetch_ui.show()

    def finish_fetch_data(self,op_tag,input_name,src_ui,idx=None):
        if idx is None:
            idx = src_ui.source_treeview.currentIndex()
        op = self.paw.get_op(op_tag)
        src = op.input_locator[input_name].src 
        if src == optools.wf_input:
            v = str(src_ui.source_treeview.model().get_uri_of_index(idx)).strip()
        elif src == optools.fs_input:
            v = str(src_ui.source_treeview.model().filePath(idx)).strip()
        self.paw.set_input(op_tag,input_name,val=v)
        #dest_widget.setText(v)
        src_ui.close()

    #def finish_edit_op(self,op_tag,val_widgets,ui):
    #    op = self.paw.get_op(op_tag)
    #    for input_name in op.inputs.keys():
    #        self.paw.set_input(op_tag,input_name,val=val_widgets[input_name].text())
    #    ui.close()
    #    ui.deleteLater()

    def refresh_op_list(self,op_list_model):
        loaded_op_tags = self.paw.list_op_tags(self._wfname)
        for op_row,op_name in zip(range(len(self.ops.keys())),self.ops.keys()):
            if self.deps[op_name] is not None:
                if all(self.op_tag_match(nm,loaded_op_tags) for nm in self.deps[op_name]):
                    op_list_model.set_enabled(op_row)
                else:
                    op_list_model.set_disabled(op_row)

    @staticmethod
    def op_tag_match(op_name,op_tags):
        if isinstance(op_tags,str):
            op_tags = [op_tags]
        for op_tag in op_tags:
            if len(op_tag) >= len(op_name):
                if op_name == op_tag[:len(op_name)]:
                    return True
        return False

    def latest_op_tag(self,op_name):
        for op_tag in self.paw.list_op_tags(self._wfname):
            if self.op_tag_match(op_name,op_tag):
                return op_tag
        return None

    def default_op_setup(self,op_tag):
        if self.op_tag_match('read_PONI',op_tag):
            self.paw.set_input(op_tag,'poni_file',
            src='filesystem',tp='path',val='')

        elif self.op_tag_match('read_background',op_tag):
            self.paw.set_input(op_tag,'path',
            src='filesystem',tp='path',val='')

        elif self.op_tag_match('read_image',op_tag):
            self.paw.set_input(op_tag,'path',
            src='filesystem',tp='path',val='')

        elif self.op_tag_match('calibrate_background',op_tag):
            bg_op_tag = self.latest_op_tag('read_background')
            poni_op_tag = self.latest_op_tag('read_PONI')
            self.paw.set_input(op_tag,'image_data',
            src='workflow',tp='reference',val=bg_op_name+'.outputs.image_data')
            self.paw.set_input(op_tag,'poni_dict',
            src='workflow',tp='reference',val=poni_op_name+'.outputs.poni_dict')

        elif self.op_tag_match('calibrate_image',op_tag):
            img_op_name = self.latest_op_tag('read_image') 
            poni_op_name = self.latest_op_tag('read_PONI') 
            self.paw.set_input(op_tag,'image_data',
            src='workflow',tp='reference',val=img_op_name+'.outputs.image_data')
            self.paw.set_input(op_tag,'poni_dict',
            src='workflow',tp='reference',val=poni_op_name+'.outputs.poni_dict')

        elif self.op_tag_match('subtract_background',op_tag):
            imgcal_op_name = self.latest_op_tag('calibrate_image') 
            bgcal_op_name = self.latest_op_tag('calibrate_background') 
            self.paw.set_input(op_tag,'I',
            src='workflow',tp='reference',val=imgcal_op_name+'.outputs.I')
            self.paw.set_input(op_tag,'I_bg',
            src='workflow',tp='reference',val=bgcal_op_name+'.outputs.I')

        elif self.op_tag_match('saxs_profile',op_tag):
            bgsub_op_name = self.latest_op_tag('subtract_background')
            self.paw.set_input(op_tag,'q',
            src='workflow',tp='reference',val=bgsub_op_name+'.outputs.q')
            self.paw.set_input(op_tag,'I',
            src='workflow',tp='reference',val=bgsub_op_name+'.outputs.I_at_q')
            
    def rm_op(self,op_name):
        print 'remove {}'.format(op_name)

    def set_visualizer(self,op_tag,state):
        if not state==0:
            # Find, create, or otherwise open the widget
            if not op_tag in self.output_widgets.keys():
                # Get the interesting piece of data from this op
                output_data = self.get_output_data(op_tag)
                # Form a widget from the output data 
                widg = widgets.make_widget(output_data)
                self.output_widgets[op_tag] = widg
            else:
                # The user has closed a tab
                # instead of unchecking the visualizer box
                widg = self.output_widgets[op_tag]
            if self.ui.output_tabs.indexOf(widg) == -1:
                # Add a tab to self.ui.output_tabs 
                tab_idx = self.ui.output_tabs.addTab(widg,op_tag) 
            self.ui.output_tabs.setCurrentWidget(widg)
        else:
            widg = self.output_widgets.pop(op_tag)
            if widg is not None:
                tab_idx = self.ui.output_tabs.indexOf(widg)  
                widg.close() 
                if not tab_idx == -1: 
                    self.ui.output_tabs.removeTab(tab_idx)

    def update_visuals(self):
        for widg in self.output_widgets:
            if isinstance(widg,QtGui.QWidget):
                widg.repaint()

    def get_output_data(self,op_tag):
        if self.op_tag_match('read_PONI',op_tag):
            return self.paw.get_output(op_tag,'poni_dict',self._wfname)
        elif ( self.op_tag_match('read_background',op_tag)
        or self.op_tag_match('read_image',op_tag) ):
            return self.paw.get_output(op_tag,'image_data',self._wfname)
        elif ( self.op_tag_match('calibrate_background') 
        or self.op_tag_match('calibrate_image',op_tag) ):
            return self.paw.get_output(op_tag,'q_I',self._wfname)
        elif self.op_tag_match('subtract_background',op_tag):
            return self.paw.get_output(op_tag,'q_I',self._wfname)
        elif self.op_tag_match('saxs_profile',op_tag):
            return self.paw.get_output(op_tag,'results',self._wfname)
                
    def run_wf(self):
        self.paw.execute()
        self.update_visuals()



import numpy as np
import copy

from scipy.optimize import curve_fit

from ...Operation import Operation
from ... import optools
from . import saxstools

class SpectrumFit(Operation):
    """
    Use a measured SAXS spectrum (I(q) vs. q),
    to optimize the parameters of a theoretical SAXS spectrum
    for one or several populations of scatterers. 
    Works by minimizing an objective function that compares
    the measured spectrum against the theoretical result.
    TODO: document the algorithm here.

    Input arrays of q and I(q), 
    a string indicating choice of objective function, 
    a dict of features describing the spectrum,
    and a list of strings indicating which keys in the dict 
    should be used as optimization parameters.
    The input features dict includes initial fit parameters
    as well as the flags indicating which populations to include.
    The features dict is of the same format as
    SpectrumProfiler and SpectrumParameterization outputs.

    Outputs a return code and the features dict,
    with entries updated for the optimized parameters.
    Also returns the theoretical result for I(q),
    and a renormalized measured spectrum for visual comparison.
    """

    def __init__(self):
        input_names = ['q','I','features','fit_params','objfun']
        output_names = ['return_code','features','q_I_opt']
        super(SpectrumFit, self).__init__(input_names, output_names)

        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['features'] = str('dict of features describing the input spectrum and scatterer populations. '
        + 'Keys of this dict are the same as the SpectrumParameterization.outputs.features dict.')
        self.input_doc['fit_params'] = str('List of strings indicating which parameters to optimize. '
        + 'Each of these strings should also be a key of the input features dict.')
        self.input_doc['objfun'] = str('String indicating choice of objective function for optimization. '
        + 'See documentation of saxstools.saxs_fit() for supported objective functions.')
        #self.input_doc['constraints'] = str('List of strings indicating optimization constraints. '
        #+ 'See documentation of saxstools.fit_saxs() for supported constraints.')

        self.output_doc['return_code'] = str('integer indicating whether or not the spectrum was found to be fittable. '
        + 'Possible values: 0=success, 1=error, 2=warning')
        self.output_doc['features'] = str('dict of features copied from inputs, '
        + 'with entries optimized for all keys specified in fit_params.')
        self.output_doc['q_I_opt'] = str('n-by-2 array of q and the theoretical intensity spectrum, '
        + 'normalized so that I(q=0) is near 1.') 

        self.input_src['q'] = optools.wf_input
        self.input_src['I'] = optools.wf_input
        self.input_src['objfun'] = optools.text_input
        self.input_src['fit_params'] = optools.text_input
        self.input_src['features'] = optools.wf_input
        self.input_type['q'] = optools.ref_type
        self.input_type['I'] = optools.ref_type
        self.input_type['objfun'] = optools.str_type
        self.input_type['fit_params'] = optools.str_type
        self.input_type['features'] = optools.ref_type
        # TODO: establish a good default here.
        self.inputs['objfun'] = 'full_spectrum_chi2log' 

    def run(self):
        # Set return code to 1 (error) by default;
        # if execution finishes, set it to 0
        self.outputs['return_code'] = 1
        q, I = self.inputs['q'], self.inputs['I']
        m = self.inputs['objfun']
        p = self.inputs['fit_params']
        f = copy.deepcopy(self.inputs['features'])
        if f['bad_data_flag']:
            # sabotage
            self.outputs['return_code'] = 2
            f['FIT_ERROR'] = '[{}] found bad_data flag, aborting'.format(__name__)
            self.outputs['features'] = f
            return
        c = []
        if f['form_flag'] or f['structure_flag']:
            c = ['fix_I0']
        #c = self.inputs['constraints'] 
        #try:
        #print f
        d_opt = saxstools.saxs_fit(q,I,m,f,p,c)
        f.update(d_opt)
        #print d_opt
            #for k in p:
            #    f[k] = d_opt[k]
        #except Exception as ex:
        #    msg = str('[{}] optimization by {} failed '
        #    + 'while attempting to fit spectrum. message: '
        #    + '{}'.format(__name__,m,ex.message))
        #    raise ex

        I_opt = saxstools.compute_saxs(q,f)
        q_I_opt = np.array([q,I_opt]).T

        self.outputs['features'] = f 
        self.outputs['q_I_opt'] = q_I_opt
        self.outputs['return_code'] = 0




import os

import numpy as np
import pyFAI

from ...Operation import Operation
from ... import optools

class NikaToPONI(Operation):
    """
    Converts Nika calibration output 
    (saved in a text file)
    to a dict of PyFAI PONI parameters,
    by first converting from Nika to Fit2D,
    then using a pyFAI.AzimuthalIntegrator 
    to convert from Fit2D to PONI format.

    WARNING: the map from Nika's horizontal and vertical tilts
    to Fit2D's tilt and tiltPlanRotation
    has not yet been verified by the developers. 
    Use this operation with nonzero tilts at your own risk.
 
    Input a text file expressing results of Nika automated calibration,
    and manually input polarization factor. 
    Output a dict of pyFAI PONI calibration parameters.
    Format of text file for Nika output is expected to be:
    sample_to_CCD_mm=____
    pixel_size_x_mm=____
    pixel_size_y_mm=____
    beam_center_x_pix=____
    beam_center_y_pix=____
    horizontal_tilt_deg=____
    vertical_tilt_deg=____ 
    wavelength_A=____ 
    """
    
    def __init__(self):
        input_names = ['nika_file','fpolz']
        output_names = ['poni_dict']
        super(NikaToPONI,self).__init__(input_names,output_names)
        self.input_doc['nika_file'] = str('text file expressing nika automated calibration results- '
        + 'see Operation documentation for the expected format of this file')
        self.input_doc['fpolz'] = 'polarization factor'
        self.input_src['nika_file'] = optools.fs_input
        self.input_src['fpolz'] = optools.text_input
        self.input_type['nika_file'] = optools.path_type
        self.input_type['fpolz'] = optools.float_type
        self.inputs['fpolz'] = 0.95 
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters, as found in a .poni file'

    def run(self):
        fpath = self.inputs['nika_file']
        fpolz = self.inputs['fpolz']
        for line in open(fpath,'r'):
            kv = line.strip().split('=')
            if kv[0] == 'sample_to_CCD_mm':
                d_mm = float(kv[1])         # Nika reports direct detector distance, 
                                            # from sample to where beam axis intersects detector plane, in mm.
            if kv[0] == 'pixel_size_x_mm':
                pxsz_x_mm = float(kv[1])    # Nika uses pixel dimensions in mm- this is the 'horzontal' dimension. 
            if kv[0] == 'pixel_size_y_mm':
                pxsz_y_mm = float(kv[1])    # Nika uses pixel dimensions in mm- this is the 'vertical' dimension. 
            if kv[0] == 'beam_center_x_pix':
                bcx_px = float(kv[1])       # Nika reports the x coord relative to 'bottom left' corner of detector
                                            # where beam axis intersects detector plane, in pixels 
            if kv[0] == 'beam_center_y_pix':
                bcy_px = float(kv[1])       # same as beam_center_x_pix but for y 
            if kv[0] == 'horizontal_tilt_deg':    
                htilt_deg = float(kv[1])    # Nika reports the horizontal tilt in degrees...
            if kv[0] == 'vertical_tilt_deg':
                vtilt_deg = float(kv[1])    # Nika reports the vertical tilt in degrees...
            if kv[0] == 'wavelength_A':
                wl_A = float(kv[1])         # Nika reports wavelength is in Angstroms
        # get wavelength in m 
        wl_m = wl_A*1E-10
        pxsz_x_um = pxsz_x_mm * 1000
        pxsz_y_um = pxsz_y_mm * 1000
        # TODO: figure out how Fit2D angles correspond to Nika angles:
        # cannot use this for tilted geometries until this is done.
        tilt_deg = 0
        rot_fit2d = 0
        # use a pyFAI.AzimuthalIntegrator() to do the conversion
        p = pyFAI.AzimuthalIntegrator(wavelength = wl_m) 
        p.setFit2D(d_mm,bcx_px,bcy_px,tilt_deg,rot_fit2d,pxsz_x_um,pxsz_y_um)
        poni_dict = p.getPyFAI()
        poni_dict['fpolz'] = fpolz
        self.outputs['poni_dict'] = poni_dict 

